@@ -1,7 +1,6 @@
 """Application base class.
 """
 
-import argparse
 import codecs
 import inspect
 import locale
@@ -10,8 +9,10 @@ import logging.handlers
 import os
 import sys
 
+from cliff import argparse
 from .complete import CompleteCommand
 from .help import HelpAction, HelpCommand
+from .utils import damerau_levenshtein, COST
 
 # Make sure the cliff library has a logging handler
 # in case the app developer doesn't set up logging.
@@ -51,6 +52,9 @@ class App(object):
     :param interactive_app_factory: callable to create an
                                     interactive application
     :paramtype interactive_app_factory: cliff.interactive.InteractiveApp
+    :param deferred_help: True - Allow subcommands to accept --help with
+                          allowing to defer help print after initialize_app
+    :paramtype deferred_help: bool
     """
 
     NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
@@ -121,6 +125,7 @@ class App(object):
         parser = argparse.ArgumentParser(
             description=description,
             add_help=False,
+            allow_abbrev=False,
             **argparse_kwargs
         )
         parser.add_argument(
@@ -128,32 +133,33 @@ class App(object):
             action='version',
             version='%(prog)s {0}'.format(version),
         )
-        parser.add_argument(
+        verbose_group = parser.add_mutually_exclusive_group()
+        verbose_group.add_argument(
             '-v', '--verbose',
             action='count',
             dest='verbose_level',
             default=self.DEFAULT_VERBOSE_LEVEL,
             help='Increase verbosity of output. Can be repeated.',
         )
+        verbose_group.add_argument(
+            '-q', '--quiet',
+            action='store_const',
+            dest='verbose_level',
+            const=0,
+            help='Suppress output except warnings and errors.',
+        )
         parser.add_argument(
             '--log-file',
             action='store',
             default=None,
             help='Specify a file to log output. Disabled by default.',
         )
-        parser.add_argument(
-            '-q', '--quiet',
-            action='store_const',
-            dest='verbose_level',
-            const=0,
-            help='suppress output except warnings and errors',
-        )
         if self.deferred_help:
             parser.add_argument(
                 '-h', '--help',
                 dest='deferred_help',
                 action='store_true',
-                help="show this help message and exit",
+                help="Show help message and exit.",
             )
         else:
             parser.add_argument(
@@ -161,13 +167,13 @@ class App(object):
                 action=HelpAction,
                 nargs=0,
                 default=self,  # tricky
-                help="show this help message and exit",
+                help="Show this help message and exit.",
             )
         parser.add_argument(
             '--debug',
             default=False,
             action='store_true',
-            help='show tracebacks on errors',
+            help='Show tracebacks on errors.',
         )
         return parser
 
@@ -221,6 +227,18 @@ class App(object):
             self.options, remainder = self.parser.parse_known_args(argv)
             self.configure_logging()
             self.interactive_mode = not remainder
+            if self.deferred_help and self.options.deferred_help and remainder:
+                # When help is requested and `remainder` has any values disable
+                # `deferred_help` and instead allow the help subcommand to
+                # handle the request during run_subcommand(). This turns
+                # "app foo bar --help" into "app help foo bar". However, when
+                # `remainder` is empty use print_help_if_requested() to allow
+                # for an early exit.
+                # Disabling `deferred_help` here also ensures that
+                # print_help_if_requested will not fire if called by a subclass
+                # during its initialize_app().
+                self.options.deferred_help = False
+                remainder.insert(0, "help")
             self.initialize_app(remainder)
             self.print_help_if_requested()
         except Exception as err:
@@ -287,14 +305,61 @@ class App(object):
         self.interpreter.cmdloop()
         return 0
 
+    def get_fuzzy_matches(self, cmd):
+        """return fuzzy matches of unknown command
+        """
+
+        sep = '_'
+        if self.command_manager.convert_underscores:
+            sep = ' '
+        all_cmds = [k[0] for k in self.command_manager]
+        dist = []
+        for candidate in sorted(all_cmds):
+            prefix = candidate.split(sep)[0]
+            # Give prefix match a very good score
+            if candidate.startswith(cmd):
+                dist.append((0, candidate))
+                continue
+            # Levenshtein distance
+            dist.append((damerau_levenshtein(cmd, prefix, COST)+1, candidate))
+
+        matches = []
+        match_distance = 0
+        for distance, candidate in sorted(dist):
+            if distance > match_distance:
+                if match_distance:
+                    # we copied all items with minimum distance, we are done
+                    break
+                # we copied all items with distance=0,
+                # now we match all candidates at the minimum distance
+                match_distance = distance
+            matches.append(candidate)
+
+        return matches
+
     def run_subcommand(self, argv):
         try:
             subcommand = self.command_manager.find_command(argv)
         except ValueError as err:
-            if self.options.debug:
-                raise
+            # If there was no exact match, try to find a fuzzy match
+            the_cmd = argv[0]
+            fuzzy_matches = self.get_fuzzy_matches(the_cmd)
+            if fuzzy_matches:
+                article = 'a'
+                if self.NAME[0] in 'aeiou':
+                    article = 'an'
+                self.stdout.write('%s: \'%s\' is not %s %s command. '
+                                  'See \'%s --help\'.\n'
+                                  % (self.NAME, the_cmd, article,
+                                      self.NAME, self.NAME))
+                self.stdout.write('Did you mean one of these?\n')
+                for match in fuzzy_matches:
+                    self.stdout.write('  %s\n' % match)
             else:
-                self.LOG.error(err)
+                if self.options.debug:
+                    raise
+                else:
+                    self.LOG.error(err)
             return 2
         cmd_factory, cmd_name, sub_argv = subcommand
         kwargs = {}
